@@ -96,12 +96,18 @@ ucol_safeClone(const UCollator *coll, void * /*stackBuffer*/, int32_t * pBufferS
     if (newColl == NULL) {
         *status = U_MEMORY_ALLOCATION_ERROR;
         return nullptr;
-    } else {
+    } else if (pBufferSize != NULL) {
         *status = U_SAFECLONE_ALLOCATED_WARNING;
     }
     return newColl->toUCollator();
 }
 
+U_CAPI UCollator* U_EXPORT2
+ucol_clone(const UCollator *coll, UErrorCode *status)
+{
+    return ucol_safeClone(coll, nullptr, nullptr, status);
+}
+
 U_CAPI void U_EXPORT2
 ucol_close(UCollator *coll)
 {
@@ -369,7 +375,7 @@ ucol_getStrength(const UCollator *coll)
     return ucol_getAttribute(coll, UCOL_STRENGTH, &status);
 }
 
-U_CAPI int32_t U_EXPORT2
+U_CAPI int32_t U_EXPORT2 
 ucol_getReorderCodes(const UCollator *coll,
                     int32_t *dest,
                     int32_t destCapacity,
@@ -381,7 +387,7 @@ ucol_getReorderCodes(const UCollator *coll,
     return Collator::fromUCollator(coll)->getReorderCodes(dest, destCapacity, *status);
 }
 
-U_CAPI void U_EXPORT2
+U_CAPI void U_EXPORT2 
 ucol_setReorderCodes(UCollator* coll,
                     const int32_t* reorderCodes,
                     int32_t reorderCodesLength,
@@ -393,7 +399,7 @@ ucol_setReorderCodes(UCollator* coll,
     Collator::fromUCollator(coll)->setReorderCodes(reorderCodes, reorderCodesLength, *status);
 }
 
-U_CAPI int32_t U_EXPORT2
+U_CAPI int32_t U_EXPORT2 
 ucol_getEquivalentReorderCodes(int32_t reorderCode,
                     int32_t* dest,
                     int32_t destCapacity,
